// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

#include "Bump.h"

#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <ios>
#include <iostream>
#include <iterator>
#include <map>
#include <sstream>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include "muh_curl.h"

using std::begin;
using std::cerr;
using std::end;
using std::fstream;
using std::ifstream;
using std::ios;
using std::istringstream;
using std::make_tuple;
using std::map;
using std::ofstream;
using std::pair;
using std::string;
using std::to_string;
using std::tuple;
using std::vector;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

Page::Page()
{
}

Page::Page(const std::string& site_name,
           const std::string& board_name,
           const std::string& page_name,
           const fs::path& target_dir)
    : site_nm{site_name},
      board_nm{board_name},
      page_nm{page_name},
      targ_dir{target_dir}
{
  // DESIGN: Future adaptation for sites that don't provide a JSON API
  // if (site_has_json(site_nm))
  load_json();
  // else {
  //   load_html();
  //   const auto comp_json{compile_html_to_json()};
  //   load_json(comp_json);
  // }
}

//------------------------------------------------------------------------------

// Get the live json and store it into the page db
void Page::load_json()
{
  resp("json");
  jsondb.load(raw_json);
}

//------------------------------------------------------------------------------

// Store the board's live response inside the page
void Page::resp(const std::string& page_tp)
{
  const auto page_uri{muh_curl::muh_url(site_nm, board_nm, page_nm, page_tp)};

  if (page_tp == "json")
    raw_json = muh_curl::get_resp(page_uri).str();
  else
    raw_html = muh_curl::get_resp(page_uri).str();
}

//------------------------------------------------------------------------------

// Parse the raw json into a json db
void Page::Jsondb::load(const std::string& raw_json)
{
  Json::Reader reader{};
  reader.parse(raw_json, db);
}

//------------------------------------------------------------------------------

// Get the html response, then store both to disk
void Page::wrt_json_html()
{
  resp("html");

  write_file(page_nm, "json", targ_dir, raw_json);
  write_file(page_nm, "html", targ_dir, raw_html);
}

//------------------------------------------------------------------------------

Json::Value Page::json() const
{
  return jsondb.db;
}

fs::path Page::page_dir() const
{
  return targ_dir;
}

//------------------------------------------------------------------------------

// Write various data to files
template<typename T>
void Page::write_file(const std::string& file_nm,
                      const std::string& ext,
                      const fs::path& targ_dir,
                      const T& payload)
{
  const fs::path filename{file_nm + "." + ext};
  const fs::path filepath{targ_dir / filename};
  ofstream ofs(filepath.string());
  ofs << payload;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Parameterized Bump ctor
Bump::Bump(const int argc, char** const argv)
{
  std::tie(site_nm, board_nm, force_chk) = parse_args(argc, argv);
  arch_dir = ".";
  init_arch();

  // Re-load the 'catalog' member field now that we have the needed data
  // properly available
  catalog = Page{site_nm, board_nm, "catalog", board_dir};

  find_bumps();
}

//------------------------------------------------------------------------------

// Parse the user-supplied arguments to the program
std::tuple<std::string, std::string, bool> Bump::parse_args(const int argc,
                                                            char** const argv)
{
  if (argc < 3) {
    cerr << "\nprogram argument error\n"
         << "usage: " << argv[0] << "  site_name  board_name\n";

    std::exit(-1);
  }

  const string site_nm{argv[1]};
  const string board_nm{argv[2]};

  // Optional 3rd flag argument to force checking of the archive
  bool force_chk{false};
  if (argv[3]) {
    istringstream bool_arg{argv[3]};

    if (bool_arg.str() == "true")
      bool_arg >> std::boolalpha >> force_chk;
    else if (bool_arg.str() == "1")
      bool_arg >> force_chk;
    else {
      cerr << "\n'force check' flag argument error\n"
           << "usage: 'true' or '1'\n";

      std::exit(-1);
    }
  }

  return make_tuple(site_nm, board_nm, force_chk);
}

//------------------------------------------------------------------------------

// Initialize the archive as a whole
void Bump::init_arch()
{
  std::tie(board_dir, thrds_dir) = init_arch_dirs();

  cfg_fpath = (board_dir / ".archbot.config");
  if (!fs::exists(cfg_fpath))
    ofstream cfg_ofs{cfg_fpath.string()};
}

//------------------------------------------------------------------------------

// Initialize the archive's basic directory structure
std::tuple<fs::path, fs::path> Bump::init_arch_dirs()
{
  const fs::path site_dir{site_nm};
  fs::create_directory(arch_dir / site_dir);
  const fs::path board_dir{site_dir / board_nm};
  fs::create_directory(board_dir);
  const fs::path thrds_dir{board_dir / "threads"};
  fs::create_directory(thrds_dir);
  const fs::path media_dir{board_dir / ".media"};
  fs::create_directory(media_dir);
  const fs::path thumbs_dir{board_dir / ".thumbs"};
  fs::create_directory(thumbs_dir);
  const fs::path static_dir{board_dir / ".static"};
  fs::create_directory(static_dir);

  return make_tuple(board_dir, thrds_dir);
}

//------------------------------------------------------------------------------

// Determine if the live board has any new bumps, and keep the thread IDs if so
void Bump::find_bumps()
{
  // Compare the live board's and the archive's 'last_known_bump' values
  cfg_json = rd_json_file(cfg_fpath);
  std::tie(have_bumps, last_bump) = cmp_last_bumps(catalog.json(), cfg_json);

  if (have_bumps) {
    updt_json_item(cfg_fpath, "last_known_bump", last_bump);
    catalog.wrt_json_html();

    // load up both the live catalog's & the config's bump maps
    const auto live_bump_map{parse_bump_map(catalog.json())};
    const auto cfg_bump_map{parse_bump_map(cfg_json["threads"])};

    // Compare the two bump maps, and extract out only the differing, bumped,
    // thread IDs
    bump_ids = cmp_bmp_maps(live_bump_map, cfg_bump_map);
  }
}

//------------------------------------------------------------------------------

// Loads and parses a json file
Json::Value Bump::rd_json_file(const fs::path& filepath)
{
  ifstream json_file{filepath.string()};
  return parse_json(json_file);
}

//------------------------------------------------------------------------------

// Take a raw json source and parse it into a json db
template<typename T> Json::Value Bump::parse_json(T& json_src)
{
  Json::Reader reader{};
  Json::Value json_dat{};
  reader.parse(json_src, json_dat);
  return json_dat;
}

//------------------------------------------------------------------------------

// Check the live board's 'last known bump' vs. the archive's
std::tuple<bool, std::string> Bump::cmp_last_bumps(const Json::Value& live_json,
                                                   const Json::Value& cfg_json)
{
  // We'll have to sort the 'lastBump's from the board first
  const auto curr_bumps{load_sort_bumps(live_json)};
  const auto curr_last_bump{curr_bumps[0]};  // latest bump will be first elem

  const auto prev_last_bump{cfg_json["last_known_bump"].asString()};

  if (curr_last_bump != prev_last_bump)  // We have a new bump
    return make_tuple(true, curr_last_bump);
  else
    return make_tuple(false, prev_last_bump);
}

//------------------------------------------------------------------------------

// Sort all the 'lastBump's from the live board, descending (newest first)
std::vector<std::string> Bump::load_sort_bumps(const Json::Value& live_json)
{
  vector<string> curr_bumps{};

  for (const auto& thrd : live_json)
    curr_bumps.emplace_back(thrd["lastBump"].asString());
  sort(begin(curr_bumps), end(curr_bumps), std::greater<string>());

  return curr_bumps;
}

//------------------------------------------------------------------------------

// Associate the bumps and IDs together
std::map<uint, std::string> Bump::parse_bump_map(const Json::Value& json)
{
  map<uint, string> bump_ids;

  for (const auto& thread : json)
    bump_ids.emplace(thread["threadId"].asUInt(),
                     thread["lastBump"].asString());
  return bump_ids;
}

//------------------------------------------------------------------------------

// Compare the bump maps from the live board and the archive, storing the
// difference (keep the new bumped IDs)
std::vector<uint> Bump::cmp_bmp_maps(
    const std::map<uint, std::string>& live_bump_map,
    const std::map<uint, std::string>& cfg_bump_map)
{
  vector<uint> bumped_thrds{};

  // DESIGN: This function's algorithm as a whole needs to be tested against
  // thread deletions. -191017

  // Copy the bump maps into vectors first, for easy comparison.
  const vector<pair<uint, string>> live_bump_vec{begin(live_bump_map),
                                                 end(live_bump_map)};
  vector<pair<uint, string>> cfg_bump_vec{begin(cfg_bump_map),
                                          end(cfg_bump_map)};

  // Resize the config's vector, if the live one has new threads added.
  const auto live_sz{live_bump_vec.size()};
  if (cfg_bump_vec.size() != live_sz)
    cfg_bump_vec.resize(live_sz);

  // Compare the 'lastBump's side-by-side.
  for (size_t i{0}; i < live_sz; ++i) {
    const auto& [live_id, live_bump]{live_bump_vec[i]};
    const auto& [cfg_id, cfg_bump]{cfg_bump_vec[i]};

    // DESIGN: Dealing with deletions may require confirmation of IDs as well.

    if (force_chk || (live_bump != cfg_bump))
      bumped_thrds.emplace_back(live_id);
  }

  return bumped_thrds;
}

//------------------------------------------------------------------------------

// Overwrite a value in a json file
template<typename T>
void Bump::updt_json_item(const fs::path& json_file_nm,
                          const std::string& tag_label,
                          const T& tag_dat)
{
  fstream json_iostrm{json_file_nm.string(), ios::in | ios::out};
  auto json_val{parse_json(json_iostrm)};

  json_val[tag_label] = tag_dat;

  fast_wrt_strm(json_iostrm, json_val);
}

//------------------------------------------------------------------------------

// Write json out to a file with no whitespaces in it (fastwrite)
void Bump::fast_wrt_strm(std::fstream& io_strm, const Json::Value& json)
{
  // Prepare the stream for writing
  io_strm.clear();
  io_strm.seekp(0);

  Json::FastWriter writer;
  io_strm << writer.write(json);
}

//------------------------------------------------------------------------------

std::string Bump::site() const
{
  return site_nm;
}

std::string Bump::name() const
{
  return board_nm;
}

bool Bump::has_bumps() const
{
  return have_bumps;
}

std::vector<uint> Bump::bumps() const
{
  return bump_ids;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// DESIGN: This function is too long. Refactor. -191028

// For any threads that are bumped, save all the files out, creating directories
// for each if needed.
void Bump::save_bump_thrds()
{
  // Load the config thread data
  auto cfg_thrds_dat{parse_cfg_thrds()};
  // Load the live thread data
  const auto live_thrds_dat{parse_live_thrds()};

  // DESIGN: Refactor this loop out.
  //
  // Iteratively check if each bumped thread already has a subject_directory
  // yet, and create one if not.
  for (const auto bump_id : bump_ids) {

    const auto cfg_it = cfg_thrds_dat.find(bump_id);
    if (cfg_it != cfg_thrds_dat.end()) {  // ID found, directory already created
      auto& [id, subj_dir_bmp]{*cfg_it};
      auto& [subj, dir, cfg_bmp]{subj_dir_bmp};

      // DESIGN: Refactor this out
      // Update the thread's config bump using the live bump
      const auto live_it = live_thrds_dat.find(bump_id);
      if (live_it != live_thrds_dat.end()) {
        const auto& [live_id, subj_bmp]{*live_it};
        const auto& [live_subj, live_bmp]{subj_bmp};

        cfg_bmp = live_bmp;
      }

      // Store the thread page for processing.
      bump_thrds.emplace_back(Page{site_nm, board_nm, to_string(id), dir});
    }
    //
    else {  // Bump ID not found (ie, it's a new thread), so we need to clean &
            // pad the thread's subject, then create a new directory for it.

      const auto live_it = live_thrds_dat.find(bump_id);
      if (live_it != live_thrds_dat.end()) {  // Should be true every time atp

        const auto [id, subj_bmp]{*live_it};
        const auto [subj, bmp]{subj_bmp};
        const auto dir{add_thrd_dir(subj, id, thrds_dir)};

        // Retain the new custom subj_dir name
        tuple<string, string, string> subj_dir_bmp{subj, dir, bmp};
        cfg_thrds_dat.emplace(id, subj_dir_bmp);

        // Store the thread page for processing.
        bump_thrds.emplace_back(Page{site_nm, board_nm, to_string(id), dir});
      }
    }
  }

  // Save all the bump thread's files as needed.
  for (auto& thrd : bump_thrds) {
    grab_thrd_files(thrd);
    thrd.wrt_json_html();
  }

  // DESIGN: Refactor this loop out.
  //
  // Rewrite the updated threads data back into the json, then to the config
  // file
  cfg_json["threads"].clear();
  for (const auto& e : cfg_thrds_dat) {
    const auto [id, subj_dir_bmp]{e};
    const auto [subj, dir, bmp]{subj_dir_bmp};

    Json::Value elem{};
    elem["threadId"] = id;
    elem["subject"] = subj;
    elem["subj_id_dir"] = dir;
    elem["lastBump"] = bmp;

    cfg_json["threads"].append(elem);
  }

  updt_json_item(cfg_fpath, "threads", cfg_json["threads"]);
}

//------------------------------------------------------------------------------

// Return config's thread data, including it's custom directory
std::map<uint, std::tuple<std::string, std::string, std::string>> Bump::
    parse_cfg_thrds()
{
  map<uint, tuple<string, string, string>> cfg_thrd_dat{};

  for (const auto& thrd : cfg_json["threads"]) {
    cfg_thrd_dat.emplace(thrd["threadId"].asUInt(),
                         make_tuple(thrd["subject"].asString(),
                                    thrd["subj_id_dir"].asString(),
                                    thrd["lastBump"].asString()));
  }

  return cfg_thrd_dat;
}

//------------------------------------------------------------------------------

// Return board's live thread data
std::map<uint, std::tuple<std::string, std::string>> Bump::parse_live_thrds()
{
  map<uint, tuple<string, string>> live_thrd_dat{};

  for (const auto& thrd : catalog.json()) {
    live_thrd_dat.emplace(
        thrd["threadId"].asUInt(),
        make_tuple(thrd["subject"].asString(), thrd["lastBump"].asString()));
  }

  return live_thrd_dat;
}

//------------------------------------------------------------------------------

// Given a thread's subject, create a custom directory for it under the threads
// directory
fs::path Bump::add_thrd_dir(const std::string& subj,
                            const uint id,
                            const fs::path& thrds_dir)
{
  const auto clean_subj{make_dir_name(subj)};
  const auto id_pad{make_id_pad(id)};
  const fs::path subj_pad{clean_subj + "_" + id_pad};
  const fs::path thrd_dir{thrds_dir / subj_pad};
  fs::create_directory(thrd_dir);

  return thrd_dir;
}

//------------------------------------------------------------------------------

// Clean up the thread's subject suitable for use as a portable directory name
std::string Bump::make_dir_name(const std::string& subj)
{
  string res{subj};

  const vector<pair<char, char>> bad_chars{{' ', '_'},  // intentionally first
                                           {'&', 'n'},
                                           {'?', ' '},
                                           {'!', ' '},
                                           {'^', ' '},
                                           {'$', ' '},
                                           {'*', ' '},
                                           {'=', ' '},
                                           {'/', ' '},
                                           {'\\', ' '},
                                           {'\0', ' '},
                                           {'\n', ' '},
                                           {'(', ' '},
                                           {')', ' '},
                                           {'.', ' '},
                                           {',', ' '},
                                           {':', ' '},
                                           {';', ' '}};
  for (const auto bad_char : bad_chars)
    zap_bad_char(res, bad_char);

  const vector<string> bad_strings{"nquot", "napos"};  // '&' already xformd 'n'
  for (const auto& bad_string : bad_strings)
    zap_bad_string(res, bad_string);

  zap_bad_string(res, " ");

  // Limit the length of local directory names (44 + '_12345' == 50 total)
  if (res.size() > 44)
    res = res.substr(0, 44);

  return res;
}

//------------------------------------------------------------------------------

// Replace bad chars with a good char
void Bump::zap_bad_char(std::string& in_out,
                        const std::pair<char, char> bad_good)
{
  const auto [bad, good]{bad_good};
  std::transform(begin(in_out), end(in_out), begin(in_out), [=](char ch) {
    return ch == bad ? good : ch;
  });
}

//------------------------------------------------------------------------------

// Remove bad strings
void Bump::zap_bad_string(std::string& in_out, const std::string& bad)
{
  using str_sz_t = string::size_type;
  const auto str_end{string::npos};

  str_sz_t bad_len = bad.length();
  for (str_sz_t i = in_out.find(bad); i != str_end; i = in_out.find(bad)) {
    in_out.erase(i, bad_len);
  }
}

//------------------------------------------------------------------------------

// Pad the thread ID out to n digits, using zeros
std::string Bump::make_id_pad(const uint id, const uint n)
{
  std::stringstream res{};
  res << std::setfill('0') << std::setw(n) << id;
  return res.str();
}

//------------------------------------------------------------------------------

// Store all a thread's files to it's local directory
void Bump::grab_thrd_files(const Page& thrd)
{
  const fs::path media_dir{thrd.page_dir() / ".media"};
  fs::create_directory(media_dir);

  grab_files(thrd.json()["files"], media_dir);

  const auto posts{thrd.json()["posts"]};
  for (const auto& post : posts)
    grab_files(post["files"], media_dir);
}

//------------------------------------------------------------------------------

// Save the thread's media files
uint Bump::grab_files(const Json::Value& files_arr, const fs::path media_dir)
{
  uint file_cnt{0};

  for (const auto& file : files_arr) {
    const auto orig_file{file["originalName"].asString()};
    const fs::path media_file{make_media_file(orig_file, media_dir)};

    if (!fs::exists(media_file)) {
      const auto path_of_file{file["path"].asString()};
      auto file_url{muh_curl::muh_file_url(site_nm, path_of_file)};
      auto file_resp = muh_curl::get_resp(file_url);

      ofstream ofsb(media_file, std::ios::binary);
      ofsb << file_resp.rdbuf();
      ++file_cnt;
    }
  }

  return file_cnt;
}

//------------------------------------------------------------------------------

// Ensure the file name isn't too long
fs::path Bump::make_media_file(const std::string& orig_file,
                               const fs::path& media_dir)
{
  fs::path res{};

  if (orig_file.size() >= 120) {  // Limit the length of local filenames
    auto [filename, ext]{split_file_ext(orig_file)};
    filename = filename.substr(0, 115);
    res = (media_dir / (filename + (ext != "" ? "." + ext : "")));
  }
  else
    res = (media_dir / orig_file);

  return res;
}

//------------------------------------------------------------------------------

// Separate a filename and it's extension out
std::tuple<std::string, std::string> Bump::split_file_ext(
    const std::string& filename)
{
  // Search backwards to find the final '.'
  auto last_dot_pos{filename.rfind('.')};

  if (last_dot_pos == string::npos) {  // dot not found
    return make_tuple(filename, "");
  }
  else {
    const string file{filename.substr(0, last_dot_pos)};
    const string ext{filename.substr(last_dot_pos + 1, string::npos)};
    return make_tuple(file, ext);
  }
}

// Copyright (2019)
// License (MIT) https://opensource.org/licenses/MIT
