// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

#pragma once

#include <map>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

// The older compiler has filesystem under the experimental namespace
#if __GNUC__ == 7
#  include <experimental/filesystem>
namespace fs = std::experimental::filesystem;
#elif __GNUC__ >= 8
#  include <filesystem>
namespace fs = std::filesystem;
#endif

#include <jsoncpp/json/json.h>
// #include <json/json.h>  // you may need this version instead

//------------------------------------------------------------------------------

// Class for storing individual page data; a field inside Bump class
class Page {
 public:
  Page();
  Page(const std::string& site_nm,
       const std::string& board_nm,
       const std::string& page_nm,
       const fs::path& targ_dir);

  fs::path page_dir() const;

  Json::Value json() const;

  void wrt_json_html();

 private:
  void load_json();

  void resp(const std::string& page_tp);

  template<typename T>
  void write_file(const std::string& file_nm,
                  const std::string& ext,
                  const fs::path& targ_dir,
                  const T& payload);

  std::string site_nm;
  std::string board_nm;
  std::string page_nm;
  fs::path targ_dir;

  std::string raw_json;
  std::string raw_html;

  // Internal for storing parsed json
  struct Jsondb {
    void load(const std::string& raw_json);
    Json::Value db;
  } jsondb;
};

//------------------------------------------------------------------------------

// The primary BUMP class
class Bump {
 public:
  Bump(const int argc, char** const argv);

  std::string site() const;

  std::string name() const;

  bool has_bumps() const;

  std::vector<uint> bumps() const;

  //

  void save_bump_thrds();

 private:
  //--- pre save_bump_thrds() ---

  std::tuple<std::string, std::string, bool> parse_args(const int argc,
                                                        char** const argv);
  void init_arch();

  std::tuple<fs::path, fs::path> init_arch_dirs();

  void find_bumps();

  Json::Value rd_json_file(const fs::path& filepath);

  template<typename T> Json::Value parse_json(T& json_src);

  std::tuple<bool, std::string> cmp_last_bumps(const Json::Value& live_json,
                                               const Json::Value& cfg_json);

  std::vector<std::string> load_sort_bumps(const Json::Value& live_json);

  std::map<uint, std::string> parse_bump_map(const Json::Value& json);

  std::vector<uint> cmp_bmp_maps(
      const std::map<uint, std::string>& live_bump_map,
      const std::map<uint, std::string>& cfg_bump_map);

  template<typename T>
  void updt_json_item(const fs::path& json_file_nm,
                      const std::string& tag_label,
                      const T& tag_dat);

  void fast_wrt_strm(std::fstream& io_strm, const Json::Value& json);

  std::string site_nm;
  std::string board_nm;
  bool force_chk;

  fs::path arch_dir;
  fs::path board_dir;
  fs::path thrds_dir;

  fs::path cfg_fpath;
  Json::Value cfg_json;
  Page catalog;

  bool have_bumps;
  std::string last_bump;
  std::vector<uint> bump_ids;

  //--- post save_bump_thrds() ---

  std::map<uint, std::tuple<std::string, std::string, std::string>>
  parse_cfg_thrds();

  std::map<uint, std::tuple<std::string, std::string>> parse_live_thrds();

  fs::path add_thrd_dir(const std::string& subj,
                        const uint id,
                        const fs::path& thrds_dir);

  std::string make_dir_name(const std::string& subj);

  void zap_bad_char(std::string& in_out, const std::pair<char, char> bad_good);

  void zap_bad_string(std::string& in_out, const std::string& bad);

  std::string make_id_pad(const uint id, const uint n = 5);

  void grab_thrd_files(const Page& thrd);

  uint grab_files(const Json::Value& files_arr, const fs::path media_dir);

  fs::path make_media_file(const std::string& orig_file,
                           const fs::path& media_dir);

  std::tuple<std::string, std::string> split_file_ext(
      const std::string& filename);

  std::vector<Page> bump_thrds;
};

// Copyright (2019)
// License (MIT) https://opensource.org/licenses/MIT
